//---------------------------------------------------------------------------
//
// Evj        Model the evolution of DNA binding site recognition
//
// Module:    Stats, various statistical calculations for Evj.
// Author:    Paul C. Anagnostopoulos
// Created:   3 June 2005
//
// This computer program fits the description in the U.S. Copyright Act
// of a "United States Government work." It was written as a part of
// my/our official duties as a Government employee. This means it cannot
// be copyrighted. The program is freely available for use without a
// copyright notice, and there are no restrictions on its use, now or
// subsequently. We retain no rights in the program. Please acknowledge
// the source of this code in any releases you make.
//
//---------------------------------------------------------------------------


import java.util.Arrays;


// This static class contains various statistical routines. There is no
// reason for them to be in a class, but this is Java, so there is no
// choice.

class Stats
{

  // The genomic uncertainty (Hg) of the standard 4-valued nucleotide
  // scheme is 2.0 bits/base.
  
  private static final double Hg = 2.0;
  
  // This table gives the exact values of E(Hnb) for a number of
  // binding sites from 1 to 50. These values were computed by the
  // calhnb program (www.lecb.ncifcrf.gov/~toms/delila/calhnb.html).
  
  private static final double E_Hnb_50[] = {
    0.00000,                                      // Placehold for [0].
    0.00000, 0.75000, 1.11090, 1.32399, 1.46291,
    1.55923, 1.62900, 1.68129, 1.72155, 1.75328,
    1.77879, 1.79966, 1.81699, 1.83159, 1.84403,
    1.85475, 1.86408, 1.87227, 1.87952, 1.88598,
    1.89177, 1.89699, 1.90172, 1.90604, 1.90998,
    1.91361, 1.91695, 1.92003, 1.92290, 1.92557,
    1.92805, 1.93038, 1.93256, 1.93460, 1.93652,
    1.93834, 1.94005, 1.94167, 1.94321, 1.94466,
    1.94604, 1.94736, 1.94861, 1.94980, 1.95094,
    1.95203, 1.95307, 1.95407, 1.95503, 1.95594 };


//                      Derived Statistics
//                      ------- ----------


  // This method calculates statistics that are derived from the
  // parameters and stored in the Param object. It is invoked as part
  // of the simulation setup.
  
  public static void calculateDerived (Param param)
  {
    // Calculate Rfrequency based on the number of possible binding
    // sites and the actual binding site count.

    param.Rfreq = log2(((double)param.getInt(Param.potentialSites)) /
                       param.getInt(Param.sites));

    // Calculate the sampling information.

    computeSamplingInfo(param);
    
    // Calculate the genomic uncertainty for the size of the genome.

    /// compute_e_n_for_Hg();
  }

//                      Sampling Information
//                      -------- -----------


  // Here we compute the sampling information.
  //
  //   E(Hnb) is the expected uncertainty of n sample bases, where n is
  //   the number of binding sites. E(Hnb) = Hg - e(n)
  
  private static void computeSamplingInfo (Param param)
  {
    int sites = param.getInt(Param.sites);
    
    // Either look up E(Hnb) or compute an approximation.
    
    if (sites < E_Hnb_50.length)
      param.E_Hnb = E_Hnb_50[sites];
    else
      param.E_Hnb = computeAE_Hnb(sites);
  }

  // This method calculates an approximation of E(Hnb) and var(Hnb).
  // It is based on a formula derived by Jeff Haemer. See:
  //   G. P. Basharin, "On a statistical estimate for the entropy of a
  //   sequence of independent random variables." Theory Probability
  //   Appl. 4(3): 333--336 (1959)
  
  private static double computeAE_Hnb (int sites)
  {
    double e_n = (4 - 1) / (2 * LN2 * sites);

    return Hg - e_n;
  }

//                      Rsequence
//                      ---------


  // This method calculates Rsequence for the given chromosome, using
  // the binding site positions specified in the site array.

  public static double Rsequence (Param param, int site[], byte chromosome[])
  {
    int width = param.getInt(Param.siteWidth);
    int sites = site.length;
    double Rseq = 0.0;
    int comp[] = new int[4];

    // Iterate over the positions in the bindings sites.
    
    for (int p = 0; p < width; ++p) {

      // Count the occurences of each of the four bases.
      
      Arrays.fill(comp, 0);
      for (int s = 0; s < sites; ++s) ++comp[chromosome[site[s] + p]];

      // Calculate Hs, the uncertainty of this position.
      
      double Hs = 0.0;
      for (int b = 0; b <= 3; ++b) {
        if (comp[b] > 0) {
          double freq = (double)comp[b] / sites;
          Hs -= freq * log2(freq);
        }
      }

      // Accumulate Rsequence in bits.
      
      Rseq += param.E_Hnb - Hs;
    }

    return Rseq;
  }

//                      Utilities
//                      ---------


  // The math library ought to have log2(), don't you think?
  
  private static final double LN2 = Math.log(2);

  public static double log2 (double x)
  {
    return Math.log(x) / LN2;
  }

}
